extends Node
class_name SSTVEncoder

const TWO_PI = PI * 2
const SAMPLE_RATE = 48000.0
const VIS_BIT_FREQ = {
	'ONE': 1100.0,
	'ZERO': 1300.0,
}
const PREFIX_PULSE_LENGTH = 0.1  # 100 ms
const HEADER_PULSE_LENGTH = 0.3  # 300 ms
const HEADER_BREAK_LENGTH = 0.01 # 10 ms
const VIS_BIT_LENGTH = 0.03      # 30 ms
const SYNC_PULSE_FREQ = 1200.0

var outputFloats = PackedFloat32Array()

func EncodePrefix():
	outputFloats.append_array(GenerateTone(1900, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1900, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(2300, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(2300, PREFIX_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(1500, PREFIX_PULSE_LENGTH))

func EncodeHeader(vis: Array):
	outputFloats.append_array(GenerateTone(1900, HEADER_PULSE_LENGTH))
	outputFloats.append_array(GenerateTone(SYNC_PULSE_FREQ, HEADER_BREAK_LENGTH))
	outputFloats.append_array(GenerateTone(1900, HEADER_PULSE_LENGTH))

	# VIS Code
	outputFloats.append_array(GenerateTone(SYNC_PULSE_FREQ, VIS_BIT_LENGTH))

	var parity = 0
	vis.reverse()
	for bit in vis:
		var bit_freq = VIS_BIT_FREQ["ONE"] if bit else VIS_BIT_FREQ["ZERO"]
		if bit:
			parity += 1
		outputFloats.append_array(GenerateTone(bit_freq, VIS_BIT_LENGTH))

	var parity_freq = VIS_BIT_FREQ["ZERO"] if parity % 2 == 0 else VIS_BIT_FREQ["ONE"]
	outputFloats.append_array(GenerateTone(parity_freq, VIS_BIT_LENGTH))

	outputFloats.append_array(GenerateTone(SYNC_PULSE_FREQ, VIS_BIT_LENGTH))

func GenerateTone(frequency: float, duration: float) -> PackedFloat32Array:
	var samples = PackedFloat32Array()
	var total_samples = int(duration * SAMPLE_RATE)
	
	for i in range(total_samples):
		var time = float(i) / SAMPLE_RATE
		var sample = sin(TWO_PI * frequency * time)
		samples.append(sample)
	
	return samples

func GenerateToneFromCurve(frequencies: Array, duration: float) -> PackedFloat32Array:
	var samples = PackedFloat32Array()
	var totalSamples = int(duration * SAMPLE_RATE)
	var num_points = frequencies.size()

	var phase = 0.0
	var timeStep = 1.0 / SAMPLE_RATE

	for i in range(totalSamples):
		var time = float(i) * timeStep
		var normalized_time = time / duration
		var index_f = normalized_time * (num_points - 1)
		var index = int(index_f)
		var frac = index_f - index
		var freq: float
		
		if index >= num_points - 1:
			freq = frequencies[num_points - 1]
		else:
			freq = lerp(frequencies[index], frequencies[index + 1], frac)
		
		phase += TWO_PI * freq * timeStep

		var sample = sin(phase)
		samples.append(sample)

	return samples

#func save_wav(path: String, samples: PackedFloat32Array) -> void:
	#var file = FileAccess.open(path, FileAccess.WRITE)
	#var byte_data = PackedByteArray()
#
	#var num_samples = samples.size()
	#var data_chunk_size = num_samples * 2
	#var file_size = 44 + data_chunk_size - 8
#
	## WAV Header
	#byte_data.append_array("RIFF".to_ascii_buffer())
	#byte_data.append_array(to_little_endian(file_size, 4))
	#byte_data.append_array("WAVEfmt ".to_ascii_buffer())
	#byte_data.append_array(to_little_endian(16, 4)) # PCM header size
	#byte_data.append_array(to_little_endian(1, 2))  # PCM format
	#byte_data.append_array(to_little_endian(1, 2))
	#byte_data.append_array(to_little_endian(SAMPLE_RATE, 4))
	#byte_data.append_array(to_little_endian(SAMPLE_RATE * 1 * 16 / 8, 4)) # Byte rate
	#byte_data.append_array(to_little_endian(1 * 16 / 8, 2)) # Block align
	#byte_data.append_array(to_little_endian(16, 2))
	#byte_data.append_array("data".to_ascii_buffer())
	#byte_data.append_array(to_little_endian(data_chunk_size, 4))
#
	## Sample data
	#for sample in samples:
		#var int_sample = int(clamp(sample * 32767, -32767, 32767))
		#byte_data.append_array(to_little_endian(int_sample, 2))
#
	#file.store_buffer(byte_data)
	#file.close()
#
#func to_little_endian(value: int, byte_count: int) -> PackedByteArray:
	#var ba = PackedByteArray()
	#for i in range(byte_count):
		#ba.append((value >> (8 * i)) & 0xFF)
	#return ba
